import json
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

from eth_utils import to_dict, to_hex, to_tuple
import pytest
from vyper import compiler
from web3 import Web3

from ethpm import Package
from ethpm.tools import builder as b
from ethpm.typing import Manifest
from pytest_ethereum.deployer import Deployer


@pytest.fixture
def w3() -> Web3:
    w3 = Web3(Web3.EthereumTesterProvider())
    return w3


CONTRACTS_DIR = Path("./contracts")
SOURCES_GLOB = "**/*.vy"


@pytest.fixture
def manifest() -> Manifest:
    if not CONTRACTS_DIR.is_dir():
        raise FileNotFoundError("no contracts_dir")
    all_sources = CONTRACTS_DIR.glob(SOURCES_GLOB)
    compiler_output = generate_compiler_output(all_sources)
    composed_contract_types = generate_contract_types(compiler_output)
    composed_inline_sources = generate_inline_sources(compiler_output)
    manifest = b.build(
        {},
        b.package_name("greeter"),
        b.version("1.0.0"),
        b.manifest_version("2"),
        *composed_inline_sources,
        *composed_contract_types,
        b.validate(),
    )
    return manifest


def twig_manifest(path: Path, name: str, version: str) -> Manifest:
    all_sources = path.glob(SOURCES_GLOB)
    compiler_output = generate_compiler_output(all_sources)
    composed_contract_types = generate_contract_types(compiler_output)
    composed_inline_sources = generate_inline_sources(compiler_output)
    manifest = b.build(
        {},
        b.package_name(name),
        b.version(version),
        b.manifest_version("2"),
        *composed_inline_sources,
        *composed_contract_types,
        b.validate(),
    )
    return manifest


@to_tuple
def generate_inline_sources(compiler_output: Dict[str, Any]) -> Iterable[Manifest]:
    for path in compiler_output.keys():
        contract_type = path.split("/")[-1].split(".")[0]
        yield b.inline_source(contract_type, compiler_output)


@to_tuple
def generate_contract_types(compiler_output: Dict[str, Any]) -> Iterable[Manifest]:
    for path in compiler_output.keys():
        contract_type = path.split("/")[-1].split(".")[0]
        yield b.contract_type(contract_type, compiler_output)


@to_dict
def generate_compiler_output(
    all_sources: List[Path]
) -> Iterable[Tuple[str, Dict[str, Any]]]:
    for source in all_sources:
        contract_file = str(source).split("/")[-1]
        contract_type = contract_file.split(".")[0]
        # todo fix to accomodate multiple types in a single contract file
        yield str(source), {contract_type: create_raw_asset_data(source.read_text())}


def create_raw_asset_data(source: str) -> Dict[str, Any]:
    return {
        "abi": compiler.mk_full_signature(source),
        "evm": {
            "bytecode": {
                "object": to_hex(compiler.compile(source)),
                "linkReferences": {},
            }
        },
    }


@pytest.fixture
def package(manifest: Manifest, w3: Web3) -> Package:
    return Package(manifest, w3)


# todo squash deployers
@pytest.fixture
def vy_deployer(package: Package) -> Deployer:
    return Deployer(package)


@pytest.fixture
def twig_deployer(w3: Web3) -> Callable[[Path, str, str], Deployer]:
    def _twig_deployer(
        path: Path, name: Optional[str] = "twig", version: Optional[str] = "1.0.0"
    ) -> Deployer:
        manifest = twig_manifest(path, name, version)
        pkg = Package(manifest, w3)
        return Deployer(pkg)

    return _twig_deployer


@pytest.fixture
def solc_deployer(w3: Web3) -> Callable[[Path], Deployer]:
    def _solc_deployer(path: Path) -> Deployer:
        manifest = json.loads(path.read_text())
        package = Package(manifest, w3)
        return Deployer(package)

    return _solc_deployer



from typing import Any, Dict

from eth_tester.exceptions import TransactionFailed
import pytest
from web3.contract import ContractEvent

from pytest_ethereum._utils.abi import merge_args_and_kwargs

TxReceipt = Dict[str, Any]


class Log:
    def __init__(
        self, contract_event: ContractEvent, *args: Any, **kwargs: Any
    ) -> None:
        """
        The ``Log`` class is available to help with testing for contract events,
        and the contents of the emitted logs.
        """
        self.event = contract_event()
        self.args = merge_args_and_kwargs(self.event.abi, args, kwargs=kwargs)
        self.kwargs = kwargs

    def is_present(self, receipt: TxReceipt) -> bool:
        """
        Asserts that *every* member of ``args`` / ``kwargs`` is present in the emitted
        log dictionary values.

        .. code:: python

           assert Log(ping.events.Ping, b"one").is_present(receipt)
           assert Log(ping.events.Ping, first=b"one").is_present(receipt)
           assert Log(ping.events.Ping, first=b"one", second=b"two").is_present(receipt)
           assert Log(ping.events.Ping, b"missing").is_present(receipt) is False
           assert Log(ping.events.Ping, second=b"one").is_present(receipt) is False
           assert Log(ping.events.Ping, b"one", b"missing").is_present(receipt) is False

        """
        logs = self._process_receipt(receipt)
        missing_args = [arg for arg in self.args if arg not in logs.values()]
        if missing_args:
            return False
        return True

    def not_present(self, receipt: TxReceipt) -> bool:
        """
        Asserts that *every* member of ``args`` / ``kwargs`` are *not* present in the
        emitted log dictionary values.

        .. code:: python

           assert Log(ping.events.Ping, b"missing").not_present(receipt)
           assert Log(ping.events.Ping, first=b"missing").not_present(receipt)
           assert Log(ping.events.Ping, b"one").not_present(receipt) is False
           assert Log(ping.events.Ping, first=b"one").not_present(receipt) is False
           assert Log(ping.events.Ping, b"one", b"missing").not_present(receipt) is False
        """
        logs = self._process_receipt(receipt)
        matching_args = [arg for arg in self.args if arg in logs.values()]
        if matching_args:
            return False
        return True

    def exact_match(self, receipt: TxReceipt) -> bool:
        """
        Asserts that the provided ``kwargs`` match *exactly* the emitted log dictionary.
        Requires ``**kwargs``, and does not accept ``*args``.

        .. code:: python

           assert Log(ping.events.Ping, first=b"one", second=b"two").exact_match(receipt)
           assert Log(ping.events.Ping, first=b"one").exact_match(receipt) is False
           assert Log(ping.events.Ping, first=b"not_present").exact_match(receipt) is False
        """
        if not self.kwargs:
            raise TypeError(
                "Log().exact_match() requires keyword arguments to test an exact match."
            )

        logs = self._process_receipt(receipt)

        if self.kwargs != logs:
            return False
        return True

    def _process_receipt(self, receipt: TxReceipt) -> Dict[str, bytes]:
        processed_receipt = self.event.processReceipt(receipt)[0]
        return {k: v.rstrip(b"\x00") for k, v in processed_receipt["args"].items()}


def tx_fail(*args: Any, **kwargs: Any) -> None:
    return pytest.raises(TransactionFailed, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    setup,
    find_packages,
)

extras_require = {
    'test': [
        "pytest==3.3.2",
        "pytest-xdist",
        "tox>=2.9.1,<3",
    ],
    'lint': [
        'black>=18.6b4,<19',
        "flake8==3.4.1",
        "isort>=4.2.15,<5",
        "mypy<0.600",
    ],
    'doc': [
        "Sphinx>=1.6.5,<2",
        "sphinx_rtd_theme>=0.1.9",
    ],
    'dev': [
        "bumpversion>=0.5.3,<1",
        "pytest-watch>=4.1.0,<5",
        "wheel",
        "twine",
        "ipython",
    ],
}

extras_require['dev'] = (
    extras_require['dev'] +
    extras_require['test'] +
    extras_require['lint'] +
    extras_require['doc']
)

setup(
    name='pytest-ethereum',
    # *IMPORTANT*: Don't manually change the version here. Use `make bump`, as described in readme
    version='0.1.3-alpha.3',
    description="""pytest-ethereum: Pytest library for ethereum projects.""",
    long_description_markdown_filename='README.md',
    author='pytest-ethereum-contributors',
    author_email='',
    url='https://github.com/ethereum/pytest-ethereum',
    include_package_data=True,
    install_requires=[
        "eth-abi>=1.2.2,<2",
        "eth-utils>=1,<2",
        "ethpm>=0.1.4a1,<1",
        'web3[tester]>=4.7,<5',
        "vyper>=0.1.0b2,<1",
    ],
    setup_requires=['setuptools-markdown'],
    python_requires='>=3.6, <4',
    extras_require=extras_require,
    entry_points={"pytest11": ["pytest_ethereum = pytest_ethereum.plugins"]},
    py_modules=['pytest_ethereum'],
    license="MIT",
    zip_safe=False,
    keywords='ethereum',
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)
